/*
 * PROTOTYPE
 * Part of lp:indicator-ubuntuone
 *
 * This code calculates the total amount of bytes sent/received in a second
 * for all the files transferrred therefore providing better accuracy for estimation.
 *
 * To compile:
 * valac-0.16 --pkg gee-1.0 --pkg gio-2.0 speed-estimation.vala
 *
 * Roman Yepishev <roman.yepishev@ubuntu.com>
 */

using Gee;
using GLib;

[DBus (name="com.ubuntuone.SyncDaemon.Status")]
interface Status : Object {
    public signal void upload_started(string path);
    public signal void download_started(string path);
    public signal void upload_finished(string path,
                                       HashTable<string, string> info);
    public signal void download_finished(string path,
                                       HashTable<string, string> info);
    public signal void status_changed(HashTable<string, string> info);
    [DBus (name="current_downloads")]
    public abstract HashTable<string, string>[] current_downloads() throws IOError;
    [DBus (name="current_uploads")]
    public abstract HashTable<string, string>[] current_uploads() throws IOError;
    [DBus (name="current_status")]
    public abstract HashTable<string, string> current_status() throws IOError;
}

class Application : Object {
    public signal void transfer_speed_changed();
    public uint transfer_speed = 0; 

    private Status status_if;
    private HashTable<string, uint?> transfers;

    private bool is_polling;
    /* app-specific */
    private bool needs_redraw = true;

    private const int POLL_INTERVAL = 1 * 1000;

    public Application() {
        transfers = new HashTable<string, uint?>(str_hash, str_equal);
    }

    private uint calculate_speed(HashTable<string,string>[] current_transfers_list) {
        uint total_speed = 0;

        foreach (var item in current_transfers_list) {
            var path = item["path"];
            if (!(path in transfers)) {
                continue;
            }

            uint bytes_transferred = get_bytes_transferred(item);
            int speed = 0;

            if (bytes_transferred == transfers[path]) {
                speed = 0;
            }
            else {
                speed = (int)(bytes_transferred - transfers[path]);
                if (speed < 0) {
                    speed = 0;
                }
                transfers[path] = bytes_transferred;
            }

            total_speed += (uint)speed;
        }

        return total_speed;
    }

    private uint get_bytes_transferred(HashTable<string, string> item) {
        uint result = 0;
        if ("n_bytes_written" in item) {
            result = (uint)int.parse(item["n_bytes_written"]);
        }
        else if ("n_bytes_read" in item) {
            result = (uint)int.parse(item["n_bytes_read"]);
        }
        return result;
    }

    private bool poll_current_transfers() {
        if (! is_polling) {
            return false;
        }

        HashTable<string, string>[]? uploads, downloads;

        try {
            uploads = status_if.current_uploads();
            downloads = status_if.current_downloads();
        }
        catch (Error e) {
            warning("Error while getting current transfers: %s", e.message);
            return true; // continue polling
        }

        if (uploads.length == 0 && downloads.length == 0) {
            stop_current_transfers_polling();
            return false;
        }

        uint _transfer_speed = calculate_speed(uploads) +
                               calculate_speed(downloads);

        if (_transfer_speed != transfer_speed) {
            transfer_speed = _transfer_speed;
            transfer_speed_changed();
        }

        return true;
    }

    private void start_current_transfers_polling() {
        if (! is_polling) {
            GLib.Timeout.add(POLL_INTERVAL, poll_current_transfers);
            is_polling = true;
        }
    }

    private void stop_current_transfers_polling() {
        is_polling = false;
    }

    private void add_path_to_watch(string path,
                                   HashTable<string, string>[] current_transfers) {
        uint bytes_transferred = 0;
        foreach (var item in current_transfers) {
            if (item["path"] == path) {
                bytes_transferred = get_bytes_transferred(item);
                break;
            }
        }

        transfers[path] = bytes_transferred;
    }

    private void status_changed_cb(HashTable<string,string> info) {
        /* Stop polling if we are online.
           Start polling when we become online. Since there is an early check
           in the timer callback routine this will stop the timer if there are
           not current transfers */
        bool is_online = info["is_online"] == "True" ? true : false;
        if (is_online && info["queues"] == "WORKING") {
            start_current_transfers_polling();
        }
        else {
            transfer_speed_changed();
            stop_current_transfers_polling();
        }
    }

    private void upload_started_cb(string path) {
        start_current_transfers_polling();
        try {
            var uploads = status_if.current_uploads();
            add_path_to_watch(path, uploads);
        }
        catch(Error e) {
            warning("Failed to get current uploads: %s", e.message);
        }
    }

    private void download_started_cb(string path) {
        start_current_transfers_polling();
        try {
            var downloads = status_if.current_downloads();
            add_path_to_watch(path, downloads);
        }
        catch(Error e) {
            warning("Failed to get current downloads: %s", e.message);
        }
    }

    private void upload_finished_cb(string path, HashTable<string, string> info) {
        if (! (path in transfers)) {
            return;
        }
        transfers.remove(path);
    }

    private void download_finished_cb(string path, HashTable<string, string> info) {
        if (! (path in transfers)) {
            return;
        }
        transfers.remove(path);
    }

    public void run() {
        try {
            status_if = Bus.get_proxy_sync(BusType.SESSION,
                                           "com.ubuntuone.SyncDaemon",
                                           "/status");
        }
        catch(Error e) {
            critical("Unable to get proxy for Status interface: %s",
                     e.message);
            return;
        }

        HashTable<string, string>[]? uploads = null, downloads = null;

        status_if.upload_started.connect(upload_started_cb);
        status_if.upload_finished.connect(upload_finished_cb);
        status_if.download_started.connect(download_started_cb);
        status_if.download_finished.connect(download_finished_cb);
        status_if.status_changed.connect(status_changed_cb);
        
        try {    
            downloads = status_if.current_downloads();
            uploads = status_if.current_uploads();
        }
        catch(Error e) {
            warning("Unable to get current uploads/downloads: %s",
                    e.message);
        }

        /* 
         * we don't know when these started so we will assume the current
         * time
         */
        bool request_status = false;

        if (downloads != null && downloads.length > 0) {
            foreach (var download in downloads) {
                add_path_to_watch(download["path"], downloads);
            }
            request_status = true;
        }

        if (uploads != null && uploads.length > 0) {
            foreach (var upload in uploads) {
                add_path_to_watch(upload["path"], uploads);
            }
            request_status = true;
        }

        if (request_status) {
            try {
                var state = status_if.current_status();
                status_changed_cb(state);
            }
            catch(Error e) {
                warning("Failed to retrieve current status: %s",
                        e.message);
            }
        }

        var loop = new GLib.MainLoop();
        loop.run();
    }

    public bool show_current_transfer_speed() {
        if (needs_redraw) {
            stdout.printf("Current transfer speed: %u\n", transfer_speed);
            needs_redraw = false;
        }
        return true;
    }

    public void invalidate_transfer_speed_display() {
        needs_redraw = true;
    }

    public static int main(string[] args) {
        var app = new Application();
        app.transfer_speed_changed.connect(app.invalidate_transfer_speed_display);
        GLib.Timeout.add(1000, app.show_current_transfer_speed);
        app.run();
        return 0;
    }

}
