/*
 * Copyright 2010-2012 Roman Yepishev <roman.yepishev@ubuntu.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using GLib;
using Gee;
using Gtk;
using Gdk;
using AppIndicator;

class PublicFileMenuItem : Gtk.MenuItem {
     public string public_url { get; set; }
     public string unique_id { get; set; }
}

class Indicator : GLib.Object {
    /* Real AppIndicator implementation */

    private const int MAX_RECENT_PUBLIC_FILES=10;

    public signal void status_toggle_activated();
    public signal void public_file_item_activated(string public_url,
                                                  string filename);
    public signal void open_folder_item_activated();
    public signal void open_website_item_activated();
    public signal void open_controlpanel_item_activated();

    private AppIndicator.Indicator indicator;
    private Gtk.MenuItem state_name_item;
    private Gtk.MenuItem estimate_item;
    private Gtk.MenuItem public_files_parent_item;
    private Gtk.MenuItem status_toggle_item;
    private Gtk.MenuItem quota_item;

    private Gtk.Menu public_files_menu;

    public enum IconType {
        OFFLINE,
        WORKING, 
        IDLE,
        ERROR,
    }

    public enum ConnectionState {
        UNKNOWN,
        DISCONNECTED,
        CONNECTED,
    }

    public Indicator() {
        indicator = new AppIndicator.Indicator(
                            "ubuntuone",
                            "indicator-ubuntuone-idle",
                            IndicatorCategory.OTHER);
        indicator.set_status(IndicatorStatus.ACTIVE);
        // This is the title of indicator for accessibility
        // purposes and will be displayed in HUD
        indicator.set_title(_("Ubuntu One"));

        var menu = new Gtk.Menu();

        var open_folder_item = new Gtk.MenuItem.with_label(
                // Opens main Ubuntu One folder in file browser
                _("Open Ubuntu One Folder"));
        open_folder_item.activate.connect(() => {
                this.open_folder_item_activated();
        });

        var open_website_item = new Gtk.MenuItem.with_label(
                // Opens Ubuntu One Website in a web browser
                _("Launch Ubuntu One Website"));
        open_website_item.activate.connect(() => {
                this.open_website_item_activated();
        });

        state_name_item = new Gtk.MenuItem.with_label(_("State unknown"));
        state_name_item.set_sensitive(false);

        estimate_item = new Gtk.MenuItem.with_label(_("Fetching queue info..."));
        estimate_item.set_sensitive(false);

        quota_item = new Gtk.MenuItem.with_label(_("Free space unknown"));
        quota_item.set_sensitive(false);

        public_files_parent_item = new Gtk.MenuItem.with_label(
                _("Recently Published Files"));

        public_files_menu = new Gtk.Menu();

        public_files_parent_item.submenu = public_files_menu;
        public_files_parent_item.show();
        public_files_parent_item.set_sensitive(false);

        status_toggle_item = new Gtk.MenuItem.with_label(
                                    _("Fetching status info..."));
        status_toggle_item.activate.connect(() => {
                status_toggle_activated();
        });

        var open_controlpanel_item = new Gtk.MenuItem.with_label(
                                    // Opens ubuntuone-control-panel-qt
                                    _("Open Control Panel"));
        open_controlpanel_item.activate.connect(() => {
                open_controlpanel_item_activated();
        });

        menu.append(open_folder_item);
        menu.append(open_website_item);
        menu.append(new Gtk.SeparatorMenuItem());

        menu.append(state_name_item);
        menu.append(estimate_item);

        menu.append(new Gtk.SeparatorMenuItem());
        menu.append(quota_item);
        menu.append(open_controlpanel_item);

        menu.append(new Gtk.SeparatorMenuItem());
        menu.append(public_files_parent_item);

        menu.append(new Gtk.SeparatorMenuItem());
        menu.append(status_toggle_item);

        menu.show_all();

        indicator.set_menu(menu);
    }

    public void set_state_label(string label) {
        state_name_item.set_label(label);
    }

    public void set_quota_label(string label) {
        quota_item.set_label(label);
    }
    
    public void set_estimate_label(string label) {
        estimate_item.set_label(label);
    }

    public void set_connection_state(ConnectionState state) {
        var menu_enabled = false;
        string toggle_label;

        switch (state) {
            case ConnectionState.CONNECTED:
                menu_enabled = true;
                toggle_label = _("Disconnect");
                break;
            case ConnectionState.DISCONNECTED:
                menu_enabled = true;
                toggle_label = _("Connect");
                break;
            default:
                menu_enabled = false;
                toggle_label = _("Fetching status info...");
                break; // seriously, vala?
        }

        status_toggle_item.set_sensitive(menu_enabled);
        status_toggle_item.set_label(toggle_label);
    }

    public void set_icon(IconType icon) {
        string icon_name = "indicator-ubuntuone-offline";
        switch (icon) {
            case IconType.WORKING:
                icon_name = "indicator-ubuntuone-updating";
                break;
            case IconType.ERROR:
                icon_name = "indicator-ubuntuone-error";
                break;
            case IconType.IDLE:
                icon_name = "indicator-ubuntuone-idle";
                break;
        };
        indicator.set_icon(icon_name);
    }

    public void add_public_file_menu_item(string unique_id, string path, string public_url) {
        var item = new PublicFileMenuItem();

        item.set_label(path);
        item.public_url = public_url;
        item.unique_id = unique_id;

        item.activate.connect((menu_item) => {
            var public_file_menu_item = (PublicFileMenuItem) menu_item;
            this.public_file_item_activated(
                public_file_menu_item.public_url,
                public_file_menu_item.get_label()
            );
        });

        GLib.List<weak Widget> children = public_files_menu.get_children();

        if (children.length() > MAX_RECENT_PUBLIC_FILES) {
            /* we simply need to pop the children list */
            Gtk.Widget? extra_item = null;

            /*
             * Ideally I'd use children.last() but it does not appear to work;
             */
            children.reverse();
            foreach (weak Gtk.Widget child in children) {
                extra_item = child;
                break;
            }

            if (extra_item != null) {
                public_files_menu.remove(extra_item);
            }
            else {
                warning("Could not get the last child.");
            }
        }
        public_files_menu.prepend(item);
        item.show();
        public_files_parent_item.set_sensitive(true);
    }

    public void remove_public_file_menu_item(string unique_id) {
        /* Iterate over all menu entries and search for unique_id */
        GLib.List<weak Widget> children = public_files_menu.get_children();
        weak Gtk.Widget victim = null;
        foreach (weak Gtk.Widget child in children) {
            string child_unique_id = ((PublicFileMenuItem)child).unique_id;
            if (child_unique_id == unique_id) {
                victim = child;
            }
        }
        if (victim != null) {
            public_files_menu.remove(victim);
        }

        children = public_files_menu.get_children();
        if (children.length() == 0) {
            // Hiding it completely makes it not discoverable
            public_files_parent_item.set_sensitive(false);
        }
    }

    public Gdk.Display get_display() {
        return indicator.get_menu().get_display();
    }
}

