/*
 * Copyright 2010-2012 Roman Yepishev <roman.yepishev@ubuntu.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using GLib;
using Gtk;
using Gdk;
using Notify;

extern void exit(int exit_code);

class Application: Object {
    const string INDICATOR_BUS_NAME = "net.lappyfamily.indicator.ubuntuone";
    const uint MAX_SIMULTANEOUS_TRANSFERS = 10;

    private Indicator indicator;
    private SyncDaemonClient client;
    private Notification notification;
    private bool needs_redraw = true;
    private bool timed_refresh_enabled = false;
    int64 old_transfer_rate = 0;
    private static Application app;

    private ulong status_signal_id;

    private HashTable <string, string> state_description;

    public Application() {
    }

    void run() {
        try {
            indicator = new Indicator();
            client = new SyncDaemonClient();
        } catch (Error e) {
            stdout.printf("An error occured during initialization: %s",
                          e.message);
            exit(1);
        }

        Notify.init("indicator-ubuntuone");
        notification = new Notification("...", "...", "ubuntuone");

        state_description = new HashTable<string, string>(str_hash, str_equal);
        state_description.insert("LOCAL_RESCAN",
                                 _("Checking local files..."));
        state_description.insert("STANDOFF",
                                 _("Finishing the transfers..."));
        state_description.insert("READY",
                                 _("Ready to connect"));
        state_description.insert("SET_CAPABILITIES",
                                 _("Negotiating connection..."));
        state_description.insert("CHECK_VERSION",
                                 _("Checking client version..."));
        state_description.insert("AUTHENTICATE",
                                 _("Authenticating..."));
        state_description.insert("AUTH_FAILED",
                                 _("Authentication failed"));
        state_description.insert("SERVER_RESCAN",
                                 _("Checking files on the server..."));
        state_description.insert("WAITING",
                                 _("Waiting before reconnect..."));

        /* By default the service is not running */
        client.syncdaemon_found.connect(() => {
                /* This is a sync call, but we will get response as
                 * status-changed signal
                 */
                client.request_status();
                /* This is an async call, we will get the response as
                 * public-file-list-received
                 */
                client.request_public_file_list();
        });

        /* This is called for both SD not running and its vanishing */
        client.syncdaemon_lost.connect(syncdaemon_missing_handler);

        client.status_changed.connect(update_status_info);
        client.public_access_changed.connect(public_access_changed_handler);
        client.public_file_list_received.connect(public_file_list_received_handler);
        client.quota_changed.connect(quota_changed_handler);
        client.pending_items_changed.connect(invalidate_estimation_label);
        client.transfer_rate_changed.connect(invalidate_estimation_label);
 
        indicator.open_folder_item_activated.connect(() => {
            try {
                Gtk.show_uri(null, client.root, Gdk.CURRENT_TIME); 
            }
            catch(Error e) {
                warning("Failed to open %s: %s", client.root, e.message);
            }
        });
        indicator.open_website_item_activated.connect(() => {
            try{
                Gtk.show_uri(null, "https://one.ubuntu.com/", Gdk.CURRENT_TIME);
            }
            catch(Error e) {
                warning("Failed to launch Ubuntu One web site: %s", e.message);
            }
        });
        indicator.open_controlpanel_item_activated.connect(() => {
            try {
                GLib.Process.spawn_command_line_async("ubuntuone-control-panel-qt");
            }
            catch(Error e) {
                warning("Failed to spawn control panel: %s", e.message);
            }
        });

        indicator.public_file_item_activated.connect(copy_public_file_url);

        /* Default handler is to attempt to connect */ 
        status_signal_id = indicator.status_toggle_activated.connect(client.connect_daemon);
        client.request_quota_info();
    }

    private void quota_changed_handler(int64 quota_total, int64 quota_used) {
        int64 quota_free = quota_total - quota_used;
        string label;
        if (quota_free < 0) {
            label = _("Over quota by %s").printf(Util.format_data(-quota_free));
        }
        else {
            label = _("Free space: %s").printf(Util.format_data(quota_free));
        }
        indicator.set_quota_label(label);
    }

    private void public_access_changed_handler(HashTable <string, string> info) {
        bool is_public = true; // PublicFilesList does not have is_public;

        if ("is_public" in info) {
            is_public = (info["is_public"] == "True" ? true : false);
        }

        string path = GLib.Path.get_basename(info["path"]);
        string public_url = info["public_url"];
        string node_id = info["node_id"];
        /* volume_id can be null for root node */
        string volume_id = info["volume_id"] == null ? "" : info["volume_id"];

        string unique_id = "%s:%s".printf(volume_id, node_id);

        if (is_public) {
            indicator.add_public_file_menu_item(unique_id, path, public_url);
        }
        else {
            /* LP:975989, no luck today */
            indicator.remove_public_file_menu_item(unique_id);
        }
    }

    private void public_file_list_received_handler(
                                        HashTable <string, string>[] files) {
        /* Usually the last files will be at the end of the list
         * But we cannot be certain about that.
         */
        int list_length = files.length - 1;
        int slice_start = 0;
        if (list_length > 10) {
            slice_start = list_length - 10;
        }

        foreach (var item in files[slice_start:list_length]) {
            public_access_changed_handler(item);
        }
    }

    private void copy_public_file_url(string public_url,
                                      string filename) {
        var display = indicator.get_display();
        var clipboard = Clipboard.get_for_display(
                display, Gdk.SELECTION_CLIPBOARD);
        clipboard.set_text(public_url, -1);

        notification.update(
            _("Public link copied"),
            // %s is the name of the published file
            _("A public link to '%s' was copied to the clipboard").printf(filename),
            "ubuntuone"
        );

        try {
            notification.show();
        }
        catch(Error e) {
            warning("Error showing notification: %s", e.message);
        }
    }

    private void syncdaemon_missing_handler() {
        indicator.set_state_label(_("Service is not running"));
        indicator.set_estimate_label(_("No queue information"));
        indicator.set_connection_state(Indicator.ConnectionState.DISCONNECTED);
        indicator.set_icon(Indicator.IconType.OFFLINE);
    }

    static void initialize() {
        app = new Application();
        app.run();
    }

    static int main(string[] args) {
        /* Try to acquire DBus name to prevent multiple copies */
        Bus.own_name(BusType.SESSION,
                     INDICATOR_BUS_NAME,
                     BusNameOwnerFlags.NONE,
                    // got bus, got name
                    null, 
                    initialize, 
                    // name lost
                    () => {
            stdout.printf("%s\n",
                     _("Ubuntu One Indicator is already running."));
            exit(0);
        });

        Gtk.init(ref args);

        Gtk.main();

        return 0;
    }

    /* status_changed_cb */
    void update_status_info(HashTable <string, string> status) {
        /* Updates current status info (first line) of the indicator and
         * Connect/Disconnect menu items */
        string state = status["name"];
        string queues = status["queues"];

        bool is_working = false;
        bool is_connected = (status["is_connected"] == "True" ? true : false);
        bool is_error = (status["is_error"] == "True" ? true : false );
        bool is_idle = false;

        bool old_timed_refresh_enabled = timed_refresh_enabled;
        timed_refresh_enabled = false;

        /* If the state maps directly to the description, this is easy */
        if (state in state_description) {
            indicator.set_state_label(state_description[state]);
        }

        if (state == "QUEUE_MANAGER") {
            if (queues == "IDLE") {
                is_idle = true;
                indicator.set_state_label(_("Synchronization complete"));
                indicator.set_estimate_label(_("Your files are up to date"));
            }
            else {
                is_working = true;
                timed_refresh_enabled = true;
                invalidate_estimation_label();
                indicator.set_state_label(_("Synchronizing..."));
                if (! old_timed_refresh_enabled) {
                    GLib.Timeout.add(1000, update_queue_size_display);
                }
            }
        }

        invalidate_estimation_label();
        update_queue_size_display();

        if (is_error) {
            indicator.set_icon(Indicator.IconType.ERROR);
        }
        else if (is_working) {
            indicator.set_icon(Indicator.IconType.WORKING);
        }
        else if (is_idle) {
            indicator.set_icon(Indicator.IconType.IDLE);
        }
        else {
            indicator.set_icon(Indicator.IconType.OFFLINE);
        }

        indicator.disconnect(status_signal_id);

        if (is_connected) {
            indicator.set_connection_state(Indicator.ConnectionState.CONNECTED);
            status_signal_id = indicator.status_toggle_activated.connect(
                                    client.disconnect_daemon);
        }
        else {
            indicator.set_connection_state(Indicator.ConnectionState.DISCONNECTED);
            status_signal_id = indicator.status_toggle_activated.connect(
                                    client.connect_daemon);
        }

    }

    /* GLib.Timeout callback, so returns true if need to repeat */
    private bool update_queue_size_display() {
        if (! needs_redraw) {
            return true;
        }

        int64 eta_raw = 0;
        int64 remaining_bytes_raw = 0;
        uint files_transferring = 0; 

        bool is_online = client.is_online;
        bool is_idle = client.is_idle;
        // transfer estimation is more or less exact
        // i.e. we have only Uploads
        bool is_exact = !client.has_downloads;
        bool has_pending_bytes = client.pending_bytes > 0;

        uint ops_count = client.pending_operations;
        bool has_pending_ops = ops_count > 0;

        int64 transfer_rate_raw = (client.transfer_rate + old_transfer_rate) / 2;

        old_transfer_rate = transfer_rate_raw;

        string transfer_rate = Util.format_transfer_rate(transfer_rate_raw);

        if (is_exact) {
            files_transferring = client.pending_file_transfers;
            remaining_bytes_raw = (int64)(client.pending_bytes - 
                                          client.files_in_transfer_bytes_done);
        }
        else {
            if (client.files_in_transfer_count > 0 &&
                    client.files_in_transfer_count < client.MAX_SIMULTANEOUS_TRANSFERS-1) {

                /* 
                 * There are only <10 transfers, so it means that's everything
                 * we got, we can estimate based on that
                 */
                remaining_bytes_raw = (int64)(client.files_in_transfer_bytes_total -
                                              client.files_in_transfer_bytes_done);
                files_transferring = client.files_in_transfer_count;
            }
            else {
                // we are not sure about the estimation at all
                files_transferring = client.pending_file_transfers;
                remaining_bytes_raw = 0;
            }
        }

        remaining_bytes_raw = remaining_bytes_raw.clamp(0, int64.MAX);

        string remaining_bytes = Util.format_data(remaining_bytes_raw);
        // context: "ETA: unknown"
        string eta = _("unknown time");

        if (remaining_bytes_raw > 0 && transfer_rate_raw > 0) {
            eta_raw = remaining_bytes_raw / transfer_rate_raw;
            eta = Util.format_time(eta_raw);
        }

        string label = null;

        if (is_online) {
            if (is_idle) {
                label = _("Your files are up to date");
            }
            else {
                if (transfer_rate_raw == 0) {
                    if (files_transferring > 0) {
                        // %u: Number of files being currently in the queue
                        label = ngettext("Transferring %u file",
                                         "Transferring %u files",
                                         files_transferring).printf(files_transferring);
                    }
                    else {
                        // %u: Number of tasks we are working on
                        label = ngettext("Working on %u task...",
                                         "Working on %u tasks...",
                                         ops_count).printf(ops_count);
                    }
                }
                else {
                   if (eta_raw > 0) {
                        // %1$s: formatted display of the transfer rate
                        // %2$u: formatted display of ETA: HH:MM:SS
                        // %3$u: number of files being currently transferred
                        label = ngettext("Transferring %3$u file (%1$s, %2$s left)",
                            "Transferring %3$u files (%1$s, %2$s left)",
                            files_transferring).printf(
                                transfer_rate, eta, files_transferring);
                    }
                    else {
                        // %1$s: formatted display of the transfer rate
                        // %2$u: number of files being currently transferred
                        label = ngettext(
                            "Transferring %2$u file (%1$s)",
                            "Transferring %2$u files (%1$s)",
                            files_transferring).printf(
                                eta, files_transferring);
                    }
                }
            }
        }
        else {
            if (is_idle) {
                label = _("No pending tasks");
            }
            else {
                // %2$u - number of operations pending
                // %1$s - formatted display of the total bytes to transfer
                label = ngettext("%2$u task pending (%1$s total)",
                                 "%2$u tasks pending (%1$s total)",
                                 ops_count).printf(remaining_bytes, ops_count);
            }
        }

        indicator.set_estimate_label(label);

        if (! timed_refresh_enabled) {
            return false;
        }

        return true;
    }

    public void invalidate_estimation_label() {
        needs_redraw = true;
    }

}

