/*
 * Copyright 2010-2012 Roman Yepishev <roman.yepishev@ubuntu.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using GLib;
using Gee;
using Soup;
using OAuth;
using Json;
// using SoupGNOME;

[DBus (name = "com.ubuntuone.SyncDaemon.Status")]
interface Status : GLib.Object {
    public signal void content_queue_changed();
    public signal void meta_queue_changed();
    public signal void upload_started(string path);
    public signal void upload_finished(string path,
                                      HashTable<string, string> info);
    public signal void download_started(string path);
    public signal void download_finished(string path,
                                      HashTable<string, string> info);
    public signal void status_changed(HashTable<string, string> info);
    public signal void request_queue_added(
            string op_name, string op_id, HashTable<string,string> data);
    public signal void request_queue_removed(
            string op_name, string op_id, HashTable<string,string> data);

    [DBus (name = "current_status")]
    public abstract HashTable<string, string> current_status() throws IOError;
    [DBus (name="current_downloads")]
    public abstract HashTable<string, string>[] current_downloads() throws IOError;
    [DBus (name="current_uploads")]
    public abstract HashTable<string, string>[] current_uploads() throws IOError;

}

[DBus (name = "com.ubuntuone.SyncDaemon.PublicFiles")]
interface PublicFiles : GLib.Object {
    public signal void public_access_changed(
                                HashTable <string, string> file_info);
    public signal void public_files_list (
                                HashTable<string, string>[] files);

    [DBus (name = "get_public_files")]
    public abstract void get_public_files() throws IOError;
}

[DBus (name = "com.ubuntuone.SyncDaemon.SyncDaemon")]
interface Daemon : GLib.Object {
    [DBus (name = "connect")]
    public abstract void connect_daemon() throws IOError;
    [DBus (name = "disconnect")]
    public abstract void disconnect_daemon() throws IOError;
    // Unused, but may be used in the future
    //[DBus (name = "quit")]
    //public abstract void quit_daemon() throws IOError;
    [DBus (name = "get_rootdir")]
    public abstract string get_rootdir() throws IOError;
}

[DBus (name = "com.ubuntuone.CredentialsManagement")]
interface CredentialsManagement : GLib.Object {
    public signal void credentials_found(HashTable <string, string> info);
    public signal void credentials_error();

    [DBus (name = "find_credentials")]
    public abstract void find_credentials() throws IOError;
}

class SyncDaemonClient : GLib.Object {
    // Default value
    public const int MAX_SIMULTANEOUS_TRANSFERS=10;

    /* A simple Proxy for real SyncDaemon service */
    private const string QUOTA_URL = "https://one.ubuntu.com/api/quota/";
    private const int QUOTA_REFRESH_INTERVAL = 10 * 1000; // 10 seconds
    private const int POLL_INTERVAL = 1 * 1000;

    /* Signals we emit possibly after post-processing */
    public signal void content_queue_changed();
    public signal void meta_queue_changed();
    public signal void status_changed(HashTable <string, string> info);
    public signal void public_access_changed(
                                     HashTable<string, string> file_info);
    public signal void upload_started(string path);
    public signal void upload_finished(string path,
                                      HashTable<string, string> info);
    public signal void download_started(string path);
    public signal void download_finished(string path,
                                         HashTable<string, string> info);
    public signal void quota_info_received(HashTable<string, string> info);
    public signal void public_file_list_received(
                                        HashTable<string, string>[] files);

    /* Could not come up with better names */
    public signal void syncdaemon_found();
    public signal void syncdaemon_lost();

    /* At some point we will start sending these */
    public signal void quota_changed(int64 quota_total, int64 quota_used);

    public signal void credentials_found();
    public signal void credentials_error();

    public signal void pending_items_changed();
    public signal void transfer_rate_changed();

    public bool is_idle = false;
    public bool is_online = false;

    public int64 transfer_rate = 0; 

    public uint pending_operations = 0;
    public uint pending_file_transfers = 0;
    public int64 pending_bytes = 0;
    public bool has_downloads = false;

    /* Files simultaneously transferring */
    public uint files_in_transfer_count = 0;
    public int64 files_in_transfer_bytes_total = 0;
    public int64 files_in_transfer_bytes_done = 0;

    private Status status_if;
    private PublicFiles publicfiles_if;
    private Daemon daemon_if;
    private CredentialsManagement credentials_if;
    private Soup.SessionAsync http_session;
    private bool is_running = false;

    private int64 quota_used = 0;
    private int64 quota_total = 0;

    private bool have_credentials = false;
    private string consumer_key;
    private string consumer_secret;
    private string token_key;
    private string token_secret;

    private bool quota_refresh_requested = false;
    private bool is_polling_current_transfers;

    public string root { get; set; }
    private HashTable<string, int64?> transfers = null;
    private HashMap<string, int?> operations = null;

    public SyncDaemonClient() throws IOError {
        status_if = Bus.get_proxy_sync(BusType.SESSION,
                                       "com.ubuntuone.SyncDaemon",
                                       "/status",
                                       DBusProxyFlags.DO_NOT_AUTO_START);
        publicfiles_if = Bus.get_proxy_sync(BusType.SESSION,
                                       "com.ubuntuone.SyncDaemon",
                                       "/publicfiles",
                                       DBusProxyFlags.DO_NOT_AUTO_START);
        daemon_if = Bus.get_proxy_sync(BusType.SESSION,
                                       "com.ubuntuone.SyncDaemon",
                                       "/",
                                       DBusProxyFlags.DO_NOT_AUTO_START);
        credentials_if = Bus.get_proxy_sync(BusType.SESSION,
                                       "com.ubuntuone.Credentials",
                                       "/credentials",
                                       DBusProxyFlags.DO_NOT_AUTO_START);

        status_if.content_queue_changed.connect(() => {
                content_queue_changed();
        });

        status_if.meta_queue_changed.connect(() => {
                meta_queue_changed();
        });

        publicfiles_if.public_access_changed.connect((file_info) => {
                public_access_changed(file_info);
        });

        publicfiles_if.public_files_list.connect((info) => {
                public_file_list_received(info);
        });

        status_if.upload_started.connect(upload_started_cb);
        status_if.upload_finished.connect(upload_finished_cb);
        status_if.download_started.connect(download_started_cb);
        status_if.download_finished.connect(download_finished_cb);

        status_if.status_changed.connect((info) => {
                status_changed_cb(info);
                status_changed(info);
        });

        status_if.request_queue_added.connect(request_queue_added_cb);
        status_if.request_queue_removed.connect(request_queue_removed_cb);
        
        /*
         * XXX: LP:978934 makes the indicator to start on u1sdtool --quit
         * effectively restarting syncdaemon
         */
        Bus.watch_name(BusType.SESSION,
                "com.ubuntuone.SyncDaemon",
                BusNameWatcherFlags.NONE,
                () => { 
                    this.is_running = true;
                    this.request_credentials();
                    this.request_root();
                    this.syncdaemon_found();
                },
                () => {
                    this.is_running = false;
                    reset_pending_counters();
                    this.syncdaemon_lost();
                }
        );

        credentials_if.credentials_found.connect((info) => {
                consumer_key = info["consumer_key"];
                consumer_secret = info["consumer_secret"];
                token_key = info["token"];
                token_secret = info["token_secret"];
                have_credentials = true;
                credentials_found();
        });

        credentials_if.credentials_error.connect(
                                () => { credentials_error(); });

        http_session = new Soup.SessionAsync();
        http_session.user_agent = "%s/%s (libsoup)".printf(
                        Config.PACKAGE_NAME, Config.PACKAGE_VERSION);
        /* This is a reasonable default since we don't support changing the root
         * dir. This will be overwritten when SyncDaemon tells us about the
         * real root directory
         */
        root = "file://%s/Ubuntu One".printf(GLib.Environment.get_home_dir());

        transfers = new HashTable<string, uint?>(str_hash, str_equal);
        operations = new HashMap<string, int?>();

    }

    public void request_root() {
        if (! is_running) {
            return;
        }

        try {
            root = "file://" + daemon_if.get_rootdir();
        }
        catch(Error e) {
            warning("Root dir not found: %s", e.message);
        }
    }

    public void request_quota_info() {
        if (! have_credentials) {
            this.credentials_found.connect(this.request_quota_info);
            request_credentials();
            return;
        }

        if (quota_refresh_requested) {
            // already scheduled
            return;
        }

        if (! quota_refresh_requested && quota_total == 0) {
            // this is the first call, need to process immediately
            request_quota_info_impl();
            return;
        }

        quota_refresh_requested = true;

        GLib.Timeout.add(QUOTA_REFRESH_INTERVAL, request_quota_info_impl);
    }

    private bool request_quota_info_impl() {
        this.credentials_found.disconnect(this.request_quota_info);

        var url = OAuth.sign_url2(QUOTA_URL, /* postargs */ null,
                                    OAuth.Method.HMAC, "GET",
                                    consumer_key,
                                    consumer_secret,
                                    token_key,
                                    token_secret);

        var message = new Soup.Message("GET", url);

        /* Proxy support is a must but I can't make it work for now */
        //session.add_feature(new SoupGNOME.ProxyResolverGNOME());
        http_session.queue_message(message, (session, message) => {
            if (message.status_code != 200) {
                warning("Quota request failed: HTTP %u", message.status_code);
                return;
            }

            /* http://live.gnome.org/Vala/JsonSample */
            var parser = new Json.Parser();

            try {
                var result = (string) message.response_body.data;
                parser.load_from_data(result, -1);
            }
            catch (Error e) {
                warning("An error occured while parsing the response: %s",
                        e.message);
                // we cannot do much here.
                return;
            }

            Json.Object root_object = parser.get_root().get_object();
            quota_total = root_object.get_int_member("total");
            quota_used = root_object.get_int_member("used");

            quota_changed(quota_total, quota_used);
        });

        quota_refresh_requested = false;
        return false;
    }

    private void request_credentials() {
        // This is a separate process so OK to run if SD is not running
        try {
            credentials_if.find_credentials();
        }
        catch(Error e) {
            warning("Could not request creds: %s", e.message);
        }
    }

    public void request_status() {
        if (! is_running) {
            return;
        }

        try {
            var info = status_if.current_status();
            status_changed_cb(info);
            status_changed(info);
        }
        catch(Error e) {
            warning("Could not request status: %s", e.message);
        }
    }

    public void connect_daemon() {
        /* This is called directly from indicator so there
           should be nothing wrong in handling the error here */
        try {
            daemon_if.connect_daemon();
        }
        catch(Error e) {
            warning("Could not connect daemon: %s", e.message);
        }
    }

    public void disconnect_daemon() {
        try {
            daemon_if.disconnect_daemon();
        }
        catch(Error e) {
            warning("Could not disconnect daemon: %s", e.message);
        }
    }

    public void request_public_file_list() {
        if (! is_running) {
            return;
        }

        try {
            publicfiles_if.get_public_files();
        }
        catch(Error e) {
            warning("Could not request public file list: %s", e.message);
        }
    }

    private int64 calculate_transfer_rate(HashTable<string,string>[]
                                               current_transfers_list) {
        int64 total_speed = 0;

        foreach (var item in current_transfers_list) {
            var path = item["path"];
            if (!(path in transfers)) {
                continue;
            }

            int64 bytes_transferred = get_bytes_transferred(item);
            int64 old_bytes_transferred = transfers[path];

            if (bytes_transferred != transfers[path]) {
                total_speed += (bytes_transferred -
                                     transfers[path]).clamp(0, int64.MAX);
                transfers[path] = bytes_transferred;
            }
        }

        return total_speed;
    }

    private void update_bytes_already_transferred(
                    HashTable<string, string>[] current_transfers_list) {
        foreach(var item in current_transfers_list) {
            files_in_transfer_bytes_done += get_bytes_transferred(item);
            files_in_transfer_bytes_total += 
                                    int64.parse(item["deflated_size"]);
        }
    }

    private int64 get_bytes_transferred(HashTable<string, string> item) {
        int64 result = 0;
        if ("n_bytes_written" in item) {
            result = int64.parse(item["n_bytes_written"]);
        }
        else if ("n_bytes_read" in item) {
            result = int64.parse(item["n_bytes_read"]);
        }
        return result;
    }

    private bool poll_current_transfers() {
        if (! is_polling_current_transfers) {
            return false;
        }

        HashTable<string, string>[]? uploads, downloads;

        try {
            uploads = status_if.current_uploads();
            downloads = status_if.current_downloads();
        }
        catch (Error e) {
            warning("Error while getting current transfers: %s", e.message);
            return true; // continue polling
        }

        if (uploads.length > 0 || downloads.length > 0) {
            int64 _transfer_rate = calculate_transfer_rate(uploads) +
                                   calculate_transfer_rate(downloads);

            files_in_transfer_bytes_total = 0;
            files_in_transfer_bytes_done = 0;
            update_bytes_already_transferred(uploads);
            update_bytes_already_transferred(downloads);

            files_in_transfer_count = uploads.length + downloads.length;

            if (_transfer_rate != transfer_rate) {
                transfer_rate = _transfer_rate;
                transfer_rate_changed();
            }
        }
        else {
            files_in_transfer_count = 0;
        }

        return true;
    }

    private void start_current_transfers_polling() {
        if (! is_polling_current_transfers) {
            GLib.Timeout.add(POLL_INTERVAL, poll_current_transfers);
            is_polling_current_transfers = true;
        }
    }

    private void stop_current_transfers_polling() {
        is_polling_current_transfers = false;
    }

    private void add_path_to_watch(string path,
                                   HashTable<string, string>[] current_transfers) {
        int64 bytes_transferred = 0;
        foreach (var item in current_transfers) {
            if (item["path"] == path) {
                bytes_transferred = get_bytes_transferred(item);
                break;
            }
        }

        transfers[path] = bytes_transferred;
    }

    private void status_changed_cb(HashTable<string,string> info) {

        is_online = info["is_online"] == "True" ? true : false;
        is_idle = info["queues"] == "IDLE";
        
        HashTable<string,string>[] downloads = null, uploads = null;

        if (is_online && !is_idle) {
            start_current_transfers_polling();

            try {    
                downloads = status_if.current_downloads();
                uploads = status_if.current_uploads();
            }
            catch(Error e) {
                warning("Unable to get current uploads/downloads: %s",
                        e.message);
            }

            if (downloads != null && downloads.length > 0) {
                foreach (var download in downloads) {
                    add_path_to_watch(download["path"], downloads);
                }
            }

            if (uploads != null && uploads.length > 0) {
                foreach (var upload in uploads) {
                    add_path_to_watch(upload["path"], uploads);
                }
            }
        }
        else {
            transfer_rate = 0;
            transfer_rate_changed();
            stop_current_transfers_polling();
        }

        if (is_online && is_idle) {
            /* Cleanup the transfers */
            transfers.remove_all();
        }
    }

    private void upload_started_cb(string path) {
        start_current_transfers_polling();
        try {
            var uploads = status_if.current_uploads();
            add_path_to_watch(path, uploads);
        }
        catch(Error e) {
            warning("Failed to get current uploads: %s", e.message);
        }
    }

    private void download_started_cb(string path) {
        start_current_transfers_polling();
        try {
            var downloads = status_if.current_downloads();
            add_path_to_watch(path, downloads);
        }
        catch(Error e) {
            warning("Failed to get current downloads: %s", e.message);
        }
    }

    private void upload_finished_cb(string path, HashTable<string, string> info) {
        if (! (path in transfers)) {
            return;
        }
        transfers.remove(path);
    }

    private void download_finished_cb(string path,
                                      HashTable<string, string> info) {
        if (! (path in transfers)) {
            return;
        }
        transfers.remove(path);
    }

    void request_queue_added_cb(
            string op_name, string op_id,
            HashTable<string, string> data) {

        int64 byte_size = 0;
        if ((op_name == "Upload" || op_name == "Download") && "size" in data) {
            byte_size = int64.parse(data["size"]);
        }

        if (byte_size > 0) {
            pending_bytes += byte_size;
        }

        if (! operations.has_key(op_name)) {
            operations.set(op_name, 0);
        }

        operations.set(op_name, operations.get(op_name) + 1);
        process_operations_counters();
        pending_items_changed();
    }

    void request_queue_removed_cb(
        string op_name, string op_id,
        HashTable<string, string> data) {

        int64 byte_size = 0;
        if ((op_name == "Upload" || op_name == "Download") && "size" in data) {
            byte_size = int64.parse(data["size"]);
        }

        if (! operations.has_key(op_name)) {
            operations.set(op_name, 0);
        }

        if (operations.get(op_name) > 0) {
            operations.set(op_name, operations.get(op_name) - 1);
        }

        if (byte_size > 0) {
            pending_bytes -= byte_size;
        }

        byte_size = byte_size.clamp(0, int64.MAX);
        process_operations_counters();
        pending_items_changed();
    }

    void process_operations_counters() {
        if (operations.has_key("Download") && operations.get("Download") > 0) {
            has_downloads = true;    
        }
        else {
            has_downloads = false;
        }

        pending_operations = 0;
        pending_file_transfers = 0;

        foreach (var key in operations.keys) {
            if (key == "Upload" || key == "Download") {
                pending_file_transfers += operations.get(key);
            }
            pending_operations += operations.get(key);
        }
    }

    void reset_pending_counters() {
        files_in_transfer_count = 0;
        files_in_transfer_bytes_total = 0;
        files_in_transfer_bytes_done = 0;

        pending_file_transfers = 0;
        pending_operations = 0;
        pending_bytes = 0;

        transfer_rate = 0;

        operations.clear();
    }

}
