using GLib;
using Gee;

class TestIndicatorLabelsApp: GLib.Object {
    private SyncDaemonClient client;
    private bool needs_redraw = true;
    private bool timed_refresh_enabled = false;
    private int64 old_transfer_rate = 0;

    public TestIndicatorLabelsApp() {
        client = new SyncDaemonClient();
        client.status_changed.connect(update_status_info);
        client.pending_items_changed.connect(invalidate_estimation_label);
        client.transfer_rate_changed.connect(invalidate_estimation_label);
    }

    public void invalidate_estimation_label() {
        needs_redraw = true;
    }

    void update_status_info(HashTable<string, string> status) {
        string state = status["name"];
        string queues = status["queues"];

        bool old_timed_refresh_enabled = timed_refresh_enabled;
        timed_refresh_enabled = false;

        if (state == "QUEUE_MANAGER" && queues != "IDLE") {
            timed_refresh_enabled = true;
            invalidate_estimation_label();
            if (! old_timed_refresh_enabled) {
                GLib.Timeout.add(1000, update_queue_size_display);
            }
        }

        invalidate_estimation_label();
        update_queue_size_display();
    }

    private bool update_queue_size_display() {
        if (! needs_redraw) {
            return true;
        }

        int64 eta_raw = 0;
        int64 remaining_bytes_raw = 0;
        uint files_transferring = 0; 

        bool is_online = client.is_online;
        bool is_idle = client.is_idle;
        // transfer estimation is more or less exact
        // i.e. we have only Uploads
        bool is_exact = !client.has_downloads;
        bool has_pending_bytes = client.pending_bytes > 0;

        uint ops_count = client.pending_operations;
        bool has_pending_ops = ops_count > 0;

        int64 transfer_rate_raw = (client.transfer_rate + old_transfer_rate) / 2;

        old_transfer_rate = transfer_rate_raw;

        string transfer_rate = Util.format_transfer_rate(transfer_rate_raw);

        if (is_exact) {
            files_transferring = client.pending_file_transfers;
            remaining_bytes_raw = (int64)(client.pending_bytes - 
                                          client.files_in_transfer_bytes_done);
        }
        else {
            if (client.files_in_transfer_count > 0 &&
                    client.files_in_transfer_count < client.MAX_SIMULTANEOUS_TRANSFERS-1) {

                /* 
                 * There are only <10 transfers, so it means that's everything
                 * we got, we can estimate based on that
                 */
                remaining_bytes_raw = (int64)(client.files_in_transfer_bytes_total -
                                              client.files_in_transfer_bytes_done);
                files_transferring = client.files_in_transfer_count;
            }
            else {
                // we are not sure about the estimation at all
                files_transferring = client.pending_file_transfers;
                remaining_bytes_raw = 0;
            }
        }

        remaining_bytes_raw = remaining_bytes_raw.clamp(0, int64.MAX);

        string remaining_bytes = Util.format_data(remaining_bytes_raw);
        // context: "ETA: unknown"
        string eta = _("unknown time");

        if (remaining_bytes_raw > 0 && transfer_rate_raw > 0) {
            eta_raw = remaining_bytes_raw / transfer_rate_raw;
            eta = Util.format_time(eta_raw);
        }

        string label = null;

        if (is_online) {
            if (is_idle) {
                label = _("Your files are up to date");
            }
            else {
                if (transfer_rate_raw == 0) {
                    if (files_transferring > 0) {
                        // %u: Number of files being currently transferred
                        label = ngettext("⇄ %u file (stalled)",
                                         "⇄ %u files (stalled)",
                                         files_transferring).printf(files_transferring);
                    }
                    else {
                        label = ngettext("Working on %u task...",
                                         "Working on %u tasks...",
                                         ops_count).printf(ops_count);
                    }
                }
                else {
                    // %1$s: formatted display of transfer rate
                    // %2$u: number of files being currently transferred
                    // %3$u: number of files being currently transferred
                    label = ngettext(
                        "⇄ %3$u file at %1$s, %2$s left",
                        "⇄ %3$u files at %1$s, %2$s left",
                        files_transferring).printf(
                            transfer_rate, eta, files_transferring);
                }
            }
        }
        else {
            if (is_idle) {
                label = _("No pending tasks");
            }
            else {
                label = ngettext("%2$u task pending (%1$s total)",
                                 "%2$u tasks pending (%1$s total)",
                                 ops_count).printf(remaining_bytes, ops_count);
            }
        }

        stdout.printf("is_online=%s, is_idle=%s, is_exact=%s\n",
                      is_online.to_string(), is_idle.to_string(),
                      is_exact.to_string());
        stdout.printf("eta_raw=%lld = remaining_bytes_raw=%lld / transfer_rate_raw=%lld\n",
                      eta_raw, remaining_bytes_raw, transfer_rate_raw);

        stdout.printf("%s\n", label);

        if (! timed_refresh_enabled) {
            return false;
        }

        return true;
    }

    public static int main(string[] args) {
        var app = new TestIndicatorLabelsApp();
        var loop = new GLib.MainLoop();
        loop.run();
        return 0;
    }
}
